/*
 * Copyright (c) 2016 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <err.h>
#include <poll.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/if.h>
#undef __USE_MISC
#include <net/if.h>

#include "netlink.h"
#include "tunnel.h"

struct interface {
	char name[IFNAMSIZ];
	unsigned short type;
	unsigned int flags;
};

#define MAX_IFINDEX 256
struct interface *interfaces[MAX_IFINDEX];
int nmax_if = MAX_IFINDEX;

/*
 * Create the netlink socket
 */
int
netlink_socket(void)
{
	int nls;
	struct sockaddr_nl addr;

	if ((nls = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE)) == -1)
		return -1;

	memset (&addr, 0, sizeof(addr));
	addr.nl_family = AF_NETLINK ;
	addr.nl_groups = RTMGRP_LINK;

	if (bind(nls, (struct sockaddr *)&addr, sizeof(addr)) == -1)
		return -1;

	return nls;
}


/*
 * Callback for address change messages
 */
void
addr_change_cb(struct nlmsghdr *nlh)
{
	char name[IFNAMSIZ];
	struct ifaddrmsg *ifa = NLMSG_DATA (nlh);
	struct rtattr *rth;
	int rtl;

	rth = IFA_RTA (ifa);
	rtl = IFA_PAYLOAD (nlh);
	if (vflag)
		printf("%s: ", if_indextoname(ifa->ifa_index,name));
	for (;rtl && RTA_OK (rth, rtl);
	     rth = RTA_NEXT (rth,rtl)) {
		uint32_t ipaddr;

		if (vflag)
			printf("[ %d %d ] ", rth->rta_type, rth->rta_len);

		ipaddr = * ((uint32_t *)RTA_DATA(rth));
		ipaddr = htonl(ipaddr);

		if (vflag)
			printf("%X ", ipaddr);

	} /* for */
	if (vflag)
		printf("\n");
}

void
link_lower_change_cb(int fd, struct ifinfomsg *ifi)
{
	int up = ifi->ifi_flags & IFF_LOWER_UP;
	struct interface *idx;
	struct in_addr *dst;

	idx = interfaces[ifi->ifi_index];
	if (vflag)
		printf("link change on %s: %s\n",
		    idx->name, up ? "UP" : "DOWN");

	if (strcmp(idx->name, main_iface) == 0) {
		dst = &((struct sockaddr_in *)&dest_addr)->sin_addr;

		if (up) {
			if (backup)
				route_del(fd, dst, 32, &gateway_addr, 0);
			backup = 0;
		} else {
			route_add(fd, dst, 32, &gateway_addr, 0);
			backup = 1;
		}
	}
}


/*
 * Callback for link information change messages
 */
void
link_change_cb(int fd, struct ifinfomsg *ifi)
{
	struct interface *idx;

	if (ifi->ifi_index > nmax_if)
		errx(2, "index too high %d\n", ifi->ifi_index);
	idx = interfaces[ifi->ifi_index];
	if (idx == NULL) {
		/* 1st time */
		idx = malloc(sizeof(struct interface));
		if (idx == NULL)
			err(2, "malloc");
		if_indextoname(ifi->ifi_index, idx->name);
		idx->type = ifi->ifi_type;
		idx->flags = ifi->ifi_flags;
		interfaces[ifi->ifi_index] = idx;
		link_lower_change_cb(fd, ifi);
		return;
	}
	if ((idx->flags ^ ifi->ifi_flags) & IFF_LOWER_UP)
		link_lower_change_cb(fd, ifi);
	idx->flags = ifi->ifi_flags;
}

/*
 * Callback for routing change messages
 */
void
route_change_cb(struct nlmsghdr *nlh)
{
	struct  rtmsg *route_entry;
	struct  rtattr *route_attribute;
	int     route_attribute_len = 0;
	char    destination_address[32];
	char    gateway_address[32];
	char dev[32];

	/* Get the route data */
	route_entry = (struct rtmsg *) NLMSG_DATA(nlh);

	/* We are just intrested in main routing table */
	if (route_entry->rtm_table != RT_TABLE_MAIN)
		return;

	/* Get attributes of route_entry */
	route_attribute = (struct rtattr *) RTM_RTA(route_entry);

	/* Get the route atttibutes len */
	route_attribute_len = RTM_PAYLOAD(nlh);
	/* Loop through all attributes */
	for ( ; RTA_OK(route_attribute, route_attribute_len);
	      route_attribute = RTA_NEXT(route_attribute,
		  route_attribute_len)) {
		switch (route_attribute->rta_type) {
		case RTA_DST:
			/* Get the destination address */
			inet_ntop(AF_INET, RTA_DATA(route_attribute),
			    destination_address,
			    sizeof(destination_address));
			break;
		case RTA_GATEWAY:
			/* Get the gateway (Next hop) */
			inet_ntop(AF_INET, RTA_DATA(route_attribute),
			    gateway_address, sizeof(gateway_address));
			break;
		case RTA_OIF:
			if_indextoname(*(int *)RTA_DATA(route_attribute), dev);
			break;
		}
	}

	if (vflag) {
		if (nlh->nlmsg_type == RTM_DELROUTE)
			printf("Deleting route to %s via %s dev %s\n",
			    destination_address, gateway_address, dev);
		if (nlh->nlmsg_type == RTM_NEWROUTE)
			printf("Adding route to %s via %s dev %s\n",
			    destination_address, gateway_address, dev);
	}
}

/*
 * Send a message to add a route
 */
int
route_add(int fd, struct in_addr *dst, int prefix,
    struct in_addr *gw, int ifindex)
{
	/* buffer to hold the RTNETLINK request */
	struct {
		struct nlmsghdr hdr;
		struct rtmsg    rt;
		char            buf[8192];
	} req;
	struct msghdr rtnl_msg;
	struct rtattr *rtap;
	struct sockaddr_nl kernel;
	struct iovec io;
	int rtable = RT_TABLE_MAIN;
	size_t rtl;
	char buffer[32];

	if (fd < 0)
		return 0;

	if (vflag) {
		inet_ntop(AF_INET, dst, buffer, sizeof(buffer));
		printf("route add %s/%d via ", buffer, prefix);
		inet_ntop(AF_INET, gw, buffer, sizeof(buffer));
		printf("%s", buffer);
		if (ifindex > 0) {
			if_indextoname(ifindex, buffer);
			printf(" dev %s", buffer);
		}
		printf("\n");
	}
	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));
	kernel.nl_family = AF_NETLINK;

	/* compute the initial length of the service request */
	rtl = sizeof(struct rtmsg);

	rtap = (struct rtattr *)req.buf;
	rtap->rta_type = RTA_TABLE;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), &rtable, 4);
	rtl += rtap->rta_len;

	/* add first attrib:
	 * set destination IP addr and increment the
	 * RTNETLINK buffer size */
	rtap = (struct rtattr *) (((char *)rtap) + rtap->rta_len);
	rtap->rta_type = RTA_DST;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), dst, 4);
	rtl += rtap->rta_len;

	rtap = (struct rtattr *) (((char *)rtap) + rtap->rta_len);
	rtap->rta_type = RTA_GATEWAY;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), gw, 4);
	rtl += rtap->rta_len;

	if (ifindex != 0) {
		/* set ifc index */
		rtap = (struct rtattr *) (((char *)rtap) + rtap->rta_len);
		rtap->rta_type = RTA_OIF;
		rtap->rta_len = sizeof(struct rtattr) + 4;
		memcpy(((char *)rtap) + sizeof(struct rtattr), &ifindex, 4);
		rtl += rtap->rta_len;
	}

	/* setup the NETLINK header */
	req.hdr.nlmsg_len = NLMSG_LENGTH(rtl);
	req.hdr.nlmsg_type = RTM_NEWROUTE;
	req.hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_CREATE;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();

	/* setup the service header (struct rtmsg) */
	req.rt.rtm_family = AF_INET;
	req.rt.rtm_table = RT_TABLE_MAIN;
	req.rt.rtm_protocol = RTPROT_BOOT;
	req.rt.rtm_scope = RT_SCOPE_UNIVERSE;
	req.rt.rtm_type = RTN_UNICAST;
	/* set the network prefix size */
	req.rt.rtm_dst_len = prefix;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = (void *) &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * Send a message to delete a route
 */
int
route_del(int fd, struct in_addr *dst, int prefix,
    struct in_addr *gw, int ifindex)
{
	/* buffer to hold the RTNETLINK request */
	struct {
		struct nlmsghdr hdr;
		struct rtmsg    rt;
		char            buf[8192];
	} req;
	struct msghdr rtnl_msg;
	struct rtattr *rtap;
	struct sockaddr_nl kernel;
	struct iovec io;
	int rtable = RT_TABLE_MAIN;
	size_t rtl;
	char buffer[32];

	if (fd < 0)
		return 0;

	if (vflag) {
		inet_ntop(AF_INET, dst, buffer, sizeof(buffer));
		printf("route del %s/%d via ", buffer, prefix);
		inet_ntop(AF_INET, gw, buffer, sizeof(buffer));
		printf("%s", buffer);
		if (ifindex > 0) {
			if_indextoname(ifindex, buffer);
			printf(" dev %s", buffer);
		}
		printf("\n");
	}
	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));
	kernel.nl_family = AF_NETLINK;

	/* compute the initial length of the service request */
	rtl = sizeof(struct rtmsg);

	rtap = (struct rtattr *)req.buf;
	rtap->rta_type = RTA_TABLE;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), &rtable, 4);
	rtl += rtap->rta_len;

	/* set destination IP addr */
	rtap = (struct rtattr *) (((char *)rtap) + rtap->rta_len);
	rtap->rta_type = RTA_DST;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), dst, 4);
	rtl += rtap->rta_len;

	rtap = (struct rtattr *) (((char *)rtap) + rtap->rta_len);
	rtap->rta_type = RTA_GATEWAY;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), gw, 4);
	rtl += rtap->rta_len;

	if (ifindex != 0) {
		/* set ifc index and increment the size */
		rtap = (struct rtattr *) (((char *)rtap) + rtap->rta_len);
		rtap->rta_type = RTA_OIF;
		rtap->rta_len = sizeof(struct rtattr) + 4;
		memcpy(((char *)rtap) + sizeof(struct rtattr), &ifindex, 4);
		rtl += rtap->rta_len;
	}

	/* setup the NETLINK header */
	req.hdr.nlmsg_len = NLMSG_LENGTH(rtl);
	req.hdr.nlmsg_type = RTM_DELROUTE;
	req.hdr.nlmsg_flags = NLM_F_REQUEST|NLM_F_MATCH;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();

	/* setup the service header (struct rtmsg) */
	req.rt.rtm_family = AF_INET;
	req.rt.rtm_table = RT_TABLE_MAIN;
	req.rt.rtm_scope = RT_SCOPE_NOWHERE;
	/* set the network prefix size */
	req.rt.rtm_dst_len = prefix;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = (void *) &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * switch link up/down
 */
int
link_up(int fd, int ifindex, bool up)
{
	/* buffer to hold the RTNETLINK request */
	struct {
		struct nlmsghdr hdr;
		struct ifinfomsg info;
	} req;
	struct msghdr rtnl_msg;
	struct sockaddr_nl kernel;
	struct iovec io;

	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));
	kernel.nl_family = AF_NETLINK;

	/* setup the NETLINK header */
	req.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));
	req.hdr.nlmsg_type = RTM_SETLINK;
	req.hdr.nlmsg_flags = NLM_F_REQUEST|NLM_F_MATCH;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();

	/* setup the service header (struct ifinfomsg) */
	req.info.ifi_index = ifindex;
	req.info.ifi_family = AF_INET;
	req.info.ifi_type = 0;
	req.info.ifi_flags = up ? IFF_UP : 0;
	req.info.ifi_change = IFF_UP;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = (void *) &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * set link MTU
 */
int
link_mtu(int fd, int ifindex, int mtu)
{
	/* buffer to hold the RTNETLINK request */
	struct {
		struct nlmsghdr hdr;
		struct ifinfomsg info;
		char buf[256];
	} req;
	struct rtattr *rtap;
	struct msghdr rtnl_msg;
	struct sockaddr_nl kernel;
	struct iovec io;
	size_t rtl;

	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));
	kernel.nl_family = AF_NETLINK;

	/* compute the initial length of the service request */
	rtl = sizeof(struct ifinfomsg);

	rtap = (struct rtattr *)req.buf;
	rtap->rta_type = IFLA_MTU;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), &mtu, 4);
	rtl += rtap->rta_len;

	/* setup the NETLINK header */
	req.hdr.nlmsg_len = NLMSG_LENGTH(rtl);
	req.hdr.nlmsg_type = RTM_SETLINK;
	req.hdr.nlmsg_flags = NLM_F_REQUEST|NLM_F_MATCH;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();

	/* setup the service header (struct ifinfomsg) */
	req.info.ifi_index = ifindex;
	req.info.ifi_family = AF_INET;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = (void *) &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * Delete the tun interface
 */
int
link_delete(int fd, int ifindex)
{
	/* buffer to hold the RTNETLINK request */
	struct {
		struct nlmsghdr hdr;
		struct ifinfomsg info;
	} req;
	struct msghdr rtnl_msg;
	struct sockaddr_nl kernel;
	struct iovec io;

	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));
	kernel.nl_family = AF_NETLINK;

	/* setup the NETLINK header */
	req.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));
	req.hdr.nlmsg_type = RTM_DELLINK;
	req.hdr.nlmsg_flags = NLM_F_REQUEST|NLM_F_MATCH;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();

	/* setup the service header (struct ifinfomsg) */
	req.info.ifi_index = ifindex;
	req.info.ifi_family = AF_INET;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = (void *) &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * Setup interface addresses
 */
int
address_add(int fd, int ifindex,
    struct in_addr *local, struct in_addr *remote)
{
	/* buffer to hold the RTNETLINK request */
	struct {
		struct nlmsghdr hdr;
		struct ifaddrmsg ifa;
		char buf[256];
	} req;
	struct msghdr rtnl_msg;
	struct rtattr *rtap;
	struct sockaddr_nl kernel;
	struct iovec io;
	size_t rtl;

	if (fd < 0)
		return 0;

	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));
	kernel.nl_family = AF_NETLINK;

	/* compute the initial length of the service request */
	rtl = sizeof(struct ifaddrmsg);

	rtap = (struct rtattr *)req.buf;
	rtap->rta_type = IFA_LOCAL;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), local, 4);
	rtl += rtap->rta_len;

	rtap = (struct rtattr *) (((char *)rtap) + rtap->rta_len);
	rtap->rta_type = IFA_ADDRESS;
	rtap->rta_len = sizeof(struct rtattr) + 4;
	memcpy(((char *)rtap) + sizeof(struct rtattr), remote, 4);
	rtl += rtap->rta_len;

	/* setup the NETLINK header */
	req.hdr.nlmsg_len = NLMSG_LENGTH(rtl);
	req.hdr.nlmsg_type = RTM_NEWADDR;
	req.hdr.nlmsg_flags = NLM_F_REQUEST|NLM_F_MATCH;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();

	/* setup the service header (struct rtmsg) */
	req.ifa.ifa_family = AF_INET;
	req.ifa.ifa_prefixlen = 32;
	req.ifa.ifa_index = ifindex;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = (void *) &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * Send a message to request the list of interfaces (link level information)
 */
int
get_interfaces(int fd)
{
	struct {
		struct nlmsghdr hdr;
		struct rtgenmsg gen;
	} req;
	struct sockaddr_nl kernel;
	struct msghdr rtnl_msg;
	struct iovec io;

	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));

	kernel.nl_family = AF_NETLINK;

	req.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtgenmsg));
	req.hdr.nlmsg_type = RTM_GETLINK;
	req.hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();
	req.gen.rtgen_family = AF_PACKET;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * Send a message to request the list of addresses
 */
int
get_addresses(int fd)
{
	struct {
		struct nlmsghdr hdr;
		struct rtgenmsg gen;
	} req;
	struct sockaddr_nl kernel;
	struct msghdr rtnl_msg;
	struct iovec io;

	memset(&rtnl_msg, 0, sizeof(rtnl_msg));
	memset(&kernel, 0, sizeof(kernel));
	memset(&req, 0, sizeof(req));

	kernel.nl_family = AF_NETLINK;

	req.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtgenmsg));
	req.hdr.nlmsg_type = RTM_GETADDR;
	req.hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
	req.hdr.nlmsg_seq = 1;
	req.hdr.nlmsg_pid = getpid();
	req.gen.rtgen_family = AF_PACKET;

	io.iov_base = &req;
	io.iov_len = req.hdr.nlmsg_len;
	rtnl_msg.msg_iov = &io;
	rtnl_msg.msg_iovlen = 1;
	rtnl_msg.msg_name = &kernel;
	rtnl_msg.msg_namelen = sizeof(kernel);

	return sendmsg(fd, &rtnl_msg, 0);
}

/*
 * Callback for error messages
 */
void
error_cb(struct nlmsghdr *nlh)
{
	struct nlmsgerr *errmsg;
	struct nlmsghdr *orig;

	errmsg = (struct nlmsgerr *)NLMSG_DATA(nlh);
	orig = &errmsg->msg;
	if (vflag)
		printf("error: %d (%s)\n", orig->nlmsg_type,
		    strerror(-errmsg->error));
	switch (orig->nlmsg_type) {
	case RTM_SETLINK:
		link_change_cb(-1, (struct ifinfomsg *)NLMSG_DATA(orig));
		break;
	case RTM_NEWADDR:
		addr_change_cb(orig);
		break;
	case RTM_NEWROUTE:
	case RTM_DELROUTE:
		route_change_cb(orig);
		break;
	}

}

/*
 * Receive and dispatch netlink messages
 */
int
recv_netlink(int fd)
{
	int len;
	char reply[65536];
	struct nlmsghdr *msg_ptr;
	struct msghdr rtnl_reply;
	struct sockaddr_nl kernel;
	struct iovec io;

	memset(&io, 0, sizeof(io));
	memset(&rtnl_reply, 0, sizeof(rtnl_reply));
	memset(&kernel, 0, sizeof(kernel));
	kernel.nl_family = AF_NETLINK;
	io.iov_base = reply;
	io.iov_len = sizeof(reply);
	rtnl_reply.msg_iov = &io;
	rtnl_reply.msg_iovlen = 1;
	rtnl_reply.msg_name = &kernel;
	rtnl_reply.msg_namelen = sizeof(kernel);
	len = recvmsg(fd, &rtnl_reply, 0);

	if (len) {
		for (msg_ptr = (struct nlmsghdr *)reply;
		     NLMSG_OK(msg_ptr, len);
		     msg_ptr = NLMSG_NEXT(msg_ptr, len)) {
			if (vflag > 1)
				printf("new netlink message %d\n",
				    msg_ptr->nlmsg_type);
			switch (msg_ptr->nlmsg_type) {
			case NLMSG_DONE:
				break;
			case NLMSG_ERROR:
				error_cb(msg_ptr);
				break;
			case RTM_NEWROUTE:
			case RTM_DELROUTE:
			case RTM_GETROUTE:
				route_change_cb(msg_ptr);
				break;
			case RTM_NEWADDR:
			case RTM_DELADDR:
				addr_change_cb(msg_ptr);
				break;
			case RTM_NEWLINK:
				link_change_cb(fd, (struct ifinfomsg *)
				    NLMSG_DATA (msg_ptr));
				break;
			default:
				if (vflag)
					printf("unknown message type %d\n",
					    msg_ptr->nlmsg_type);
			} /* switch */
		}
	}
	return 0;
}
