/* ttunnel.c test tunnel Laurent GUERBY 20160411 - GNU GPLv3 */
/* OpenBSD version - Matthieu Herrb */

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <netdb.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <net/if.h>
#include <net/if_tun.h>
#include <arpa/inet.h>
#include <netinet/in.h>

int family = AF_UNSPEC;

int
addr_to_string(struct sockaddr_storage *addr, char *buffer, size_t len)
{
	switch (addr->ss_family) {
	case AF_INET:
		inet_ntop(addr->ss_family, 
		    &(((struct sockaddr_in *)addr)->sin_addr), buffer, len);
		break;
	case AF_INET6:
		inet_ntop(addr->ss_family,
		    &((struct sockaddr_in6 *)addr)->sin6_addr, buffer, len);
		break;
	default:
		snprintf(buffer, len, "unknown family");
		return -1;
	}
	return 0;
}

/*
7: tunt1: <POINTOPOINT,MULTICAST,NOARP,UP,LOWER_UP> mtu 1400 qdisc pfifo_fast state UP group default qlen 700
ip li set dev tunt1 txqueuelen 700
ip li set dev tunt1 mtu 1400
/usr/bin/nice -n -20 ./tunt
*/

int tun_alloc(char *dev_name)
{
	struct tuninfo ti;
	int fd;
	
	if( (fd = open(dev_name, O_RDWR)) < 0)
		return -1;
	memset(&ti, 0, sizeof(ti));
	if (ioctl(fd, TUNGIFINFO, &ti) < 0) {
		close(fd);
		return -1;
	}
	ti.flags |= IFF_UP;
	if (ioctl(fd, TUNSIFINFO, &ti) < 0) {
		close(fd);
		return -1;
	}
	return fd;
}

/*
 * local_socket()
 * Returns a socket listening on a local port, binds to specified source
 * address. Returns -1 on failure.
 */
int
local_socket(char *host, char *port, struct addrinfo hints)
{
	struct addrinfo *res, *res0;
	int s, ret, x = 1;
	int error;

	/* Allow nodename to be null. */
	hints.ai_flags |= AI_PASSIVE;

	/*
	 * In the case of binding to a wildcard address
	 * default to binding to an ipv4 address.
	 */
	if (host == NULL && hints.ai_family == AF_UNSPEC)
		hints.ai_family = AF_INET;

	if ((error = getaddrinfo(host, port, &hints, &res)))
		errx(1, "getaddrinfo: %s", gai_strerror(error));

	res0 = res;
	do {
		if ((s = socket(res0->ai_family, res0->ai_socktype,
		    res0->ai_protocol)) < 0)
			continue;

		ret = setsockopt(s, SOL_SOCKET, SO_REUSEPORT, &x, sizeof(x));
		if (ret == -1)
			err(1, NULL);

		if (bind(s, (struct sockaddr *)res0->ai_addr,
		    res0->ai_addrlen) == 0)
			break;

		close(s);
		s = -1;
	} while ((res0 = res0->ai_next) != NULL);

	freeaddrinfo(res);

	return (s);
}


#define IOV_N 128
#define MSG_LEN 2048

void
usage(void)
{
	errx(1, "usage: roraserv [-64knv] [-t tunX] [port]\n");
}

int main(int argc, char **argv) {
	char devname[IFNAMSIZ];
	char *uport;
	int ch, kflag = 0, nflag = 0, vflag = 0;
	int tun_fd;
	int sock;
	int result;
	struct addrinfo hints;
	struct sockaddr_storage src_addr;
	struct sockaddr_storage dest_addr;
	socklen_t src_len, dest_len;
	struct pollfd rfds[2];
	struct timespec ts1,ts2,dt,dtr;
	ssize_t count;
	ssize_t send_count;
	char msgbuf[MSG_LEN + 4];
	char buffer[256];
	socklen_t socklen;
	int prev_send_count = 0;

	snprintf(devname, sizeof(devname), "/dev/tun0");

	while ((ch = getopt(argc, argv, "46knt:v")) != -1) {
		switch (ch) {
		case '4':
			family = AF_INET;
			break;
		case '6':
			family = AF_INET6;
			break;
		case 'k':
			kflag = 1;
			break;
		case 'n':
			nflag = 1;
		case 't':
			snprintf(devname, sizeof(devname), "/dev/%s", optarg);
			break;
		case 'v':
			vflag++;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
			
	switch (argc) {
	case 0:
		uport = "30225";
		break;
	case 1:
		uport = argv[0];
		break;
	default:
		usage();
	}

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = family;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
	if (nflag) 
		hints.ai_flags |= AI_NUMERICHOST;

	sock = local_socket(NULL, uport, hints);
	if (vflag)
		printf("sock %d\n", sock);

	tun_fd = tun_alloc(devname);
	if (tun_fd < 0) 
		err(1, "tun_alloc");
	if (vflag)
		printf("tun_fd %d\n", tun_fd);

	
	rfds[0].fd = sock;
	rfds[0].events = POLLIN|POLLERR|POLLHUP;
	rfds[1].fd = tun_fd;
	rfds[1].events = POLLIN|POLLERR|POLLHUP;
	
	clock_gettime(CLOCK_MONOTONIC,&ts1);
	while (1) {
		result = poll(rfds, 2, 10000);
		
		if (result < 0) {
			err(2, "poll");
		} else if (result > 0) {
			if (rfds[0].revents) {
				uint32_t af = htonl(AF_INET);
				socklen = sizeof(struct sockaddr);
				/* room for AF header at the front */
				count = recvfrom(sock, msgbuf+4, MSG_LEN, 0,
				    (struct sockaddr *)&src_addr, &src_len);
				if (vflag > 1)
					printf("sock got %zd\n", count);
				memcpy(&dest_addr, &src_addr, src_len);
				dest_len = src_len;
				/* add AF header */
				memcpy(msgbuf, &af, 4);
				if (count > 0) {
					result = write(tun_fd, msgbuf, 
					    count + 4);
				}
			}
			if (rfds[1].revents) {
				count = read(tun_fd,msgbuf,MSG_LEN);
#if 0
				if (prev_send_count>15000) {
					clock_gettime(CLOCK_MONOTONIC,&ts2);
					dt.tv_sec = ts1.tv_sec - ts2.tv_sec;
					dt.tv_nsec = ts1.tv_nsec - ts2.tv_nsec + prev_send_count*13; /* 13 ~ 600 Mbit/s */
					if(dt.tv_sec*1000000000+dt.tv_nsec>1000) {
						nanosleep(&dt,&dtr);
					}
					ts1 = ts2;
					prev_send_count = 0;
				}
#endif
				addr_to_string(&dest_addr, buffer,
				    sizeof(buffer));
				/* strip Address family header */
				if (vflag > 1)
					printf("sock send %zd %s ", 
					    count - 4, buffer);
				send_count = sendto(sock, msgbuf+4, count-4, 0,
				    (struct sockaddr *)&dest_addr, dest_len);
				if (vflag > 1)
					printf("%zd\n", send_count);
				if (vflag && send_count < 0)
					warn("sendto");
				prev_send_count+=count;
			}
		} else {
			if (vflag > 1)
				printf("no data\n");
		}
	}
	
	return 0;
}
