/*
 * Copyright (c) 2016 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <netdb.h>
#include <poll.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <net/if.h>
#include <linux/if_tun.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "netlink.h"
#include "tunnel.h"

char *gflag = "10.68.0.5";
char *pflag;
char *sflag;
int vflag;
char *main_iface = "wan0";

int family = AF_UNSPEC;
struct sockaddr_storage dest_addr;
struct in_addr gateway_addr;
socklen_t dest_len;
int done = 0;
int tun_index;
int backup;

#define S_UDPSOCK 0
#define S_TUNNEL  1
#define S_NETLINK 2

int remote_connect(const char *, const char *, struct addrinfo);
void usage(void);

int
addr_to_string(struct sockaddr_storage *addr, char *buffer, size_t len)
{
	switch (addr->ss_family) {
	case AF_INET:
		inet_ntop(addr->ss_family,
		    &(((struct sockaddr_in *)addr)->sin_addr), buffer, len);
		break;
	case AF_INET6:
		inet_ntop(addr->ss_family,
		    &((struct sockaddr_in6 *)addr)->sin6_addr, buffer, len);
		break;
	default:
		snprintf(buffer, len, "unknown family");
		return -1;
	}
	return 0;
}

/*
 * Allocate / create the TUN interface
 */
int
tun_alloc(char *dev_name)
{
	struct ifreq ifr;
	int fd, result;

	if ((fd = open("/dev/net/tun", O_RDWR)) < 0)
		return -1;
	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
	if (*dev_name)
		strncpy(ifr.ifr_name, dev_name, IFNAMSIZ);

	if ((result = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 ) {
		close(fd);
		return result;
	}
	strncpy(dev_name, ifr.ifr_name, IFNAMSIZ);
	if (ioctl(fd, TUNSETPERSIST, 1) < 0)
		err(1, "disabling TUNSETPERSIST");

	return fd;
}

/*
 * remote_connect()
 * Returns a socket connected to a remote host. Properly binds to a local
 * port or source address if needed. Returns -1 on failure.
 */
int
remote_connect(const char *host, const char *port, struct addrinfo hints)
{
	struct addrinfo *res, *res0;
	int s, error;

	if ((error = getaddrinfo(host, port, &hints, &res)))
		errx(1, "getaddrinfo: %s", gai_strerror(error));

	res0 = res;
	do {
		if ((s = socket(res0->ai_family, res0->ai_socktype |
		    SOCK_NONBLOCK, res0->ai_protocol)) < 0)
			continue;

		/* Bind to a local port or source address if specified. */
		if (sflag || pflag) {
			struct addrinfo ahints, *ares;

			memset(&ahints, 0, sizeof(struct addrinfo));
			ahints.ai_family = res0->ai_family;
			ahints.ai_socktype = SOCK_DGRAM;
			ahints.ai_protocol = IPPROTO_UDP;
			ahints.ai_flags = AI_PASSIVE;
			if ((error = getaddrinfo(sflag, pflag, &ahints, &ares)))
				errx(1, "getaddrinfo: %s", gai_strerror(error));

			if (bind(s, (struct sockaddr *)ares->ai_addr,
			    ares->ai_addrlen) < 0)
				err(1, "bind failed");
			freeaddrinfo(ares);
		}
		/* Using unconnected UDP sockets -> no need for connect(2) */
		break;
	} while ((res0 = res0->ai_next) != NULL);

	dest_len = res0->ai_addrlen;
	memcpy(&dest_addr, res0->ai_addr, dest_len);
	char buffer[246];
	addr_to_string(&dest_addr, buffer, sizeof(buffer));
	if (vflag)
		printf("connecting to %s\n", buffer);
	freeaddrinfo(res);

	return (s);
}

void
cleanup_and_exit(int fd)
{
	struct in_addr *dst = &((struct sockaddr_in *)&dest_addr)->sin_addr;

	printf("bye\n");
	if (backup)
		route_del(fd, dst, 32, &gateway_addr, 0);
	link_up(fd, tun_index, false);
	link_delete(fd, tun_index);
}

void
ctrlc(int sig)
{
	done++;
}

void
usage(void)
{
	errx(1, "usage: roraclient [-64nv] [-g gateway] [-m iface] [-t tunx]\n"
	    "\t[-p source_port] [-s source] [destination] [port]");
}

#define IOV_N 128
#define MSG_LEN 2048

int main(int argc, char **argv) {
	char devname[IFNAMSIZ] = "roralink0";
	char *host, *uport;
	int ch, nflag = 0;
	int tun_fd, sock, nls;
	int result;
	struct addrinfo hints;
	struct sockaddr_storage src_addr;
	struct pollfd rfds[3];
	struct in_addr def;
	ssize_t count;
	char msgbuf[MSG_LEN];
	char buffer[256];
	socklen_t socklen;
	int prev_send_count = 0;
	struct in_addr local, remote;
	char local_addr[32] = "10.68.1.254", remote_addr[32] = "10.68.1.253";

	while ((ch = getopt(argc, argv, "46g:m:np:s:t:v")) != -1) {
		switch (ch) {
		case '4':
			family = AF_INET;
			break;
		case '6':
			family = AF_INET6;
			break;
		case 'g':
			gflag = optarg;
			break;
		case 'm':
			main_iface = optarg;
			break;
		case 'n':
			nflag = 1;
			break;
		case 'p':
			pflag = optarg;
			break;
		case 's':
			sflag = optarg;
			break;
		case 't':
			snprintf(devname, sizeof(devname), "%s", optarg);
			break;
		case 'v':
			vflag++;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	switch (argc) {
	case 1:
		host = argv[0];
		uport = "30225";
		break;
	case 2:
		host = argv[0];
		uport = argv[1];
		break;
	default:
		usage();
	}
	if (inet_pton(AF_INET, gflag, &gateway_addr) != 1)
		errx(2, "invalid gateway address %s", gflag);

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = family;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
	if (nflag)
		hints.ai_flags |= AI_NUMERICHOST;

	tun_fd = tun_alloc(devname);
	if (vflag)
		printf("tun_fd %d\n", tun_fd);

	sock = remote_connect(host, uport, hints);
	if (sock < 0)
		err(1, "remote_connect");

	if (vflag)
		printf("sock %d\n", sock);

	nls = netlink_socket();
	if (vflag)
		printf("netlink %d\n", nls);

	tun_index = if_nametoindex(devname);
	if (tun_index == 0)
		errx(2, "%s: not found\n", devname);

	signal(SIGINT, ctrlc);

	link_up(nls, tun_index, true);
	link_mtu(nls, tun_index, 1460);
	if (inet_pton(AF_INET, local_addr, &local) != 1)
		errx(2, "invalid local address %s", local_addr);
	if (inet_pton(AF_INET, remote_addr, &remote) != 1)
		errx(2, "invalid remote address %s", remote_addr);
	address_add(nls, tun_index, &local, &remote);

	/* set the default route through the tunnel */
	def.s_addr = INADDR_ANY;
	route_add(nls, &def, 0, &local, tun_index);

	rfds[S_UDPSOCK].fd = sock;
	rfds[S_UDPSOCK].events = POLLIN|POLLERR|POLLHUP;
	rfds[S_TUNNEL].fd = tun_fd;
	rfds[S_TUNNEL].events = POLLIN|POLLERR|POLLHUP;
	rfds[S_NETLINK].fd = nls;
	rfds[S_NETLINK].events = POLLIN|POLLERR|POLLHUP;

	get_interfaces(nls);

	while (!done) {
		result = poll(rfds, 3, 10000);

		if (result < 0) {
			warn("poll");
			break;
		} else if (result > 0) {
			if (rfds[S_UDPSOCK].revents) {
				socklen = sizeof(struct sockaddr_in);
				count = recvfrom(sock,&msgbuf,
				    MSG_LEN,0,
				    (struct sockaddr *)&src_addr, &socklen);
				if (vflag && count < 0)
					warn("recvfrom");
				if (vflag > 1)
					printf("sock got %zd\n", count);
				if (count>0) {
					result = write(tun_fd,&msgbuf,count);
				}
			}
			if (rfds[S_TUNNEL].revents) {
				count = read(tun_fd,&msgbuf,MSG_LEN);
				addr_to_string(&dest_addr, buffer,
				    sizeof(buffer));
				/* strip Address family header */
				if (vflag > 1)
					printf("sock send %zd %s ", count, buffer);
				count = sendto(sock, &msgbuf, count, 0,
				    (struct sockaddr *)&dest_addr, dest_len);
				if (vflag && count < 0)
					warn("sendto");
				if (vflag > 1)
					printf("%zd\n", count);
				prev_send_count+=count;
			}
			if (rfds[S_NETLINK].revents)
				recv_netlink(nls);
		} else {
			if (vflag > 1)
				printf("no data\n");
		}
	}
	cleanup_and_exit(nls);
	return 0;
}
